|
|  fp_fdiv: performs an FDIV operation.
|
|  The algorithm for this routine comes from Knuth's Art of Computer 
|  Programming, Volume 2 (Seminumerical Algorithms), section 4.2.3,
|  with a few addtions.  The explanation of why it works is simple 
|  and interesting; unfortunately, I forget almost all of it, and 
|  the nearest known copy of Knuth is close to an hour's drive away 
|  as I write this.  So I won't bother.  :-)
|
|  Here's an outline of the inner workings, for those scoring at home:
|  1)  Compute the exponent of the quotient:
|          quotient.exponent = dividend.exponent - divisor.exponent;
|  2)  Make sure the mantissa of the dividend is less than that of
|      the divisor.  If not, subtract the divisor mantissa from the
|      dividend mantissa.
|  3)  Use the algorithm in Knuth to compute the quotient, accurate
|      to 64 bits.
|  4)  For all rounding modes except RZ, we'll need two extra bits 
|      of accuracy: the guard bit, plus an OR of any succeeding bits 
|      (a sticky bit).  We skip to step 8 now if in RZ mode, to avoid
|      the additional overhead.
|  5)  Compute the 128-bit product of the 64-bit quotient mantissa
|      and the 64-bit divisor mantissa.  We know that the upper 64
|      bits of this product will be equal to the dividend mantissa
|      minus one; as such, we only concern ourselves with computing
|      the lower 64 bits.  Negating this result gives us the 64-bit
|      remainder.  (Confused?  Think long division.)
|  6)  If the remainder is zero, skip to step 8.  Congratulations, 
|      we have an exact result!
|  7)  Shift the divisor mantissa right one bit; compare the shifted
|      mantissa to the remainder mantissa.
|      * If the remainder mantissa is greater, set the guard and 
|        sticky bits of the quotient.
|      * If the divisor mantissa is greater, set only the sticky bit. 
|      * If the mantissas are equal: if the bit shifted off the divisor
|        mantissa was one, set only the sticky bit.  Otherwise, set
|        only the guard bit; we have an exact 65-bit result.
|  8) If we subtracted the divisor mantissa from the dividend mantissa
|     in step 2:
|         quotient.exponent++;
|         quotient.sticky |= quotient.round;
|         quotient.round = quotient.mantissa & 1LL;
|         quotient.mantissa >>= 1;
|         quotient.mantissa |= 1 << 63;
|  The routine as coded is actually a bit differently from this description;
|  still, all the important parts are correct.
|

fp_fdiv:
	|
	|  First, some preliminaries...
	|
	and.l	#0x00FF00FF, (fpe_fpsr,%a4)  | clear FPSR bytes that will be changed
	move.l	(%a0), %d0		| load source operand into %d0-%d2
	move.l	(#4,%a0), %d1
	move.l	(#8,%a0), %d2
	move.l	(%a1), %d4		| load destination operand into %d4-%d6
	move.l	(#4,%a1), %d5
	move.l	(#8,%a1), %d6
	jsr	fpd_handle_nan		| handle Not-A-Number inputs, if any

	|
	|  Compute sign - special-case results (except NAN's) always have a sign
	|  that is the exclusive OR of the two input signs (in other words, the
	|  "usual" sign).  Sign is stored in high bit of %d0.
	|
	btst	#31, %d4
	beq	1f
	bchg	#31, %d0

	|
	|  Special-case handling.
	|
1:	tst.l	%d1			| check for zero mantissa in source
	bne	chkDst			| (indicating either zero or infinity)
	tst.l	%d2
	bne	chkDst
	tst.l	%d5			| source has zero mantissa... see if dest does too
	bne	2f
	tst.l	%d6
	beq	sdzero
2:	bfextu	%d0{#16,#15}, %d7	| only source has zero mantissa...
	cmpi.w	#0x7FFF, %d7		| if source is a zero, set DZ bit in FPSR
	beq	fpd_gen_zero		| either case, return inverse of source
	or.w	#0x0410, (fpe_fpsr+2,%a4)
	bra	fpd_gen_inf
sdzero:	bfextu	%d0{#16,#15}, %d7	| dest has zero mantissa... is source infinity?
	cmpi.w	#0x7FFF, %d7
	bne	idinf
	bfextu	%d4{#16,#15}, %d7	| source is infinity... is dest infinity too?
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_zero		| if not, return a zero
	or.w	#0x2080, (fpe_fpsr+2,%a4)  | if so, set OPERR bit and return a NAN
	bra	fpd_gen_nan
idinf:	bfextu	%d4{#16,#15}, %d7	| source is zero... is dest infinity?
	cmpi.w	#0x7FFF, %d7
	beq	fpd_gen_inf		| if so, return an infinity
	or.w	#0x2080, (fpe_fpsr+2,%a4)  | if not, set OPERR bit and return a NAN
	bra	fpd_gen_nan
chkDst:	tst.l	%d5			| check for zero mantissa in dest
	bne	really			| (indicating either zero or infinity)
	tst.l	%d6			| if found, return dest
	bne	really
	bfextu	%d4{#0,#31}, %d7
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_zero
	bra	fpd_gen_inf

	|
	|  Normalize all operands, even those that would otherwise be 
	|  denormalized.  We compute the exponent, which takes up all of %d0.
	|
really:	bfextu	%d0{#16,#15}, %d3
	btst	#31, %d1
	bne	ndend
	move	%d4, -(%sp)
	tst.l	%d1
	bne	1f
	move.l	%d2, %d1
	clr.l	%d2
	sub.l	#32, %d3
1:	bfffo	%d1{#0,#32}, %d7
	move	%d7, -(%sp)
	move	%d2, -(%sp)
	neg.b	%d7
	add.b	#32, %d7
	sub.l	%d7, %d3
	lsl.l	%d7, %d1
	lsl.l	%d7, %d2
	move.l	(%sp)+, %d4 
	move.l	(%sp)+, %d7
	lsr.l	%d7, %d4
	or.l	%d4, %d1
	move.l	(%sp)+, %d4 
ndend:	bfclr	%d0{#0,#31}		| save sign in %d0 for later
	move.l	%d0, -(%sp)
	bfextu	%d4{#16,#15}, %d0	| now exponent takes up %d0
	add.w	#0x3FFF, %d0
	sub.l	%d3, %d0
	btst	#31, %d5
	bne	cdvdv
	tst.l	%d5
	bne	1f
	move.l	%d6, %d5
	clr.l	%d6
	sub.l	#32, %d0
1:	bfffo	%d1{#0,#32}, %d7
	move	%d7, -(%sp)
	move	%d6, -(%sp)
	neg.b	%d7
	add.b	#32, %d7
	sub.l	%d7, %d0
	lsl.l	%d7, %d5
	lsl.l	%d7, %d6
	move.l	(%sp)+, %d4
	move.l	(%sp)+, %d7
	lsr.l	%d7, %d4
	or.l	%d4, %d5
	| XXX - check exponent for under/overflow?

	|
	|  The actual division process starts here.
	|
cdvdv:	cmp.l	%d1, %d5		| compare dividend and divisor
	blt	ale
	bgt	presub
	cmp.l	%d2, %d6
	blt	ale
	bgt	presub
	bfclr	%d1{#0,#31}		| exact result: 2**exponent
	clr.l	%d2
	clr.l	%d3
	addq.l	#1, %d0	
	bra	eoexit
presub:	sub.l	%d6, %d2		| dividend -= divisor if the former is greater
	subx.l	%d5, %d1
	bset	#0, (%sp)		| let future code know we did this
ale:	move.l	%d2, -(%sp)		| save divisor for later
	move.l	%d1, -(%sp)
	move.l	%d2, %d4
	divu.l	%d1, %d5:%d6		| compute quotient
	clr.l	%d7
	divu.l	%d1, %d5:%d7
	mulu.l	%d6, %d3-%d2
	divu.l	%d1, %d3:%d2
	sub.l	%d2, %d7
	bcc	1f
	subq.l	#1, %d6
1:	bfextu	(fpe_fpcr,%a4){#4,#2}, %d3  | determine rounding mode
	subq.b	#1, %d3			| (cheaper than cmpi.b #1, %d0)
	beq	rmexit			| avoid lots of unnecessary work if mode is RZ
	clr.l	%d3			| compute remainder
	move.l	%d4, %d2
	mulu.l	%d7, %d3-%d2
	mulu.l	%d7, %d1
	add.l	%d1, %d3
	mulu.l	%d6, %d4
	add.l	%d4, %d3
	neg.l	%d2
	negx.l	%d3
	beq	gs00			| exact result?
	move.l	(%sp)+, %d4		| load divisor mantissa, shift one bit left
	move.l	(%sp)+, %d5
	clr.l	%d1
	lsr.l	#1, %d4
	roxr.l	#1, %d5
	roxl.l	#1, %d1			| save the bit we shifted off
	cmp.l	%d4, %d3		| now figure out guard/sticky bits
	blt	gs01e
	bgt	gs11e
	cmp.l	%d5, %d2
	blt	gs01
	bgt	gs11
	tst.l	%d1
	bne	gs01
	moveq.l	#2, %d3
	bra	gsrest
gs01e:	addq.l	#4, %sp
gs01:	moveq.l	#1, %d3
	bra	gsrest
gs11e:	addq.l	#4, %sp
gs11:	moveq.l	#3, %d3
gsrest:	ror.l	#2, %d3
rmexit:	move.l	%d7, %d2
	move.l	%d6, %d1
	btst	#0, (%sp)		| was dividend.mantissa > divisor.mantissa?
	beq	eoexit
	lsr.l	#1, %d1
	roxr.l	#1, %d2
	roxr.l	#1, %d3
	bset	#31, %d1
	addq.l	#1, %d0
	bclr	#0, (%sp)

	|
	|  Check the exponent to see if underflow/overflow occurred, or if
	|  the result must be denormalized.  If everything's OK, wrap up.
	|
eoexit:	tst.l	%d0
	bmi	minus
	cmpi.w	#0x7FFF, %d0
	blt	safe
	move.l	(%sp)+, %d0		| exponent too big - generate overflow
	bra	fpd_gen_oflow
minus:	neg.l	%d0
	cmpi.w	#64, %d0
	blt	denorm
	move.l	(%sp)+, %d0		| exponent too small - generate underflow
	bra	fpd_gen_uflow
denorm:	cmpi.w	#32, %d0
	blt	fine
	tst.l	%d3			| coarse-grained register shift
	beq	1f
	bset	#0, %d2
1:	move.l	%d2, %d3
	move.l	%d1, %d2
	clr.l	%d1
	sub.b	#32, %d0
fine:	moveq.b	#32, %d4		| fine-grained bit shift
	sub.b	%d0, %d4
	move.l	%d3, %d5
	lsr.l	%d0, %d3
	lsl.l	%d4, %d5
	beq	1f
	bset	#0, %d3
1:	move.l	%d2, %d5
	lsr.l	%d0, %d2
	lsl.l	%d4, %d5
	or.l	%d5, %d3
	move.l	%d1, %d5
	lsr.l	%d0, %d1
	lsl.l	%d4, %d5
	or.l	%d5, %d2
	clr.l	%d0
safe:	move.l	(%sp)+, %d4		| assemble high 32 bits of result
	bfins	%d0, %d4{#16,#15}
	move.l	%d4, %d0
	bra	fpd_round
