/* 
 *  arch/m68k/math-emu/fpe_composites.c
 *  (c) 1997 Rob Pelkey <rpelkey@bates.edu>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *
 *  This file implements floating-point functions that are composites of
 *  the primitive functions in the package (see primitives.c in this
 *  directory).  Many of them could be faster or more accurate if they
 *  were primitives themselves, but writing them this way cuts down on
 *  bloat, not to mention making implementing and understanding the code 
 *  less of a pain in the arse.
 *
 *  Change history:
 *  xx Dec 1997 - initial version.
 */

#include "fpe_types.h"


/*
composites remaining:
FACOS	= 2 * FATAN(FSQRT(FDIV(FADD([-1],X), FADD([1],X)))) (scale exponent)
FACOS   = FSUB([pi/2], FASIN(x)) - adjusted so that result = (-pi/2, pi/2]
FTANH	= weird algorithm, see FPSP
FATANH  = simple algorithm, see FPSP
FTWOTOX = FETOX(x * FLOGN(2))
FTENTOX = FETOX(x * FLOGN(10))
FLOG10  = FMUL(FLOGN(x), [1/ln(10)])
FLOG2   = FMUL(FLOGN(x), [1/ln(2)])
*/



/* FSUB - subtract inp2 from inp1 */
int fpe_fsub (fpe_internal *inp1, fpe_internal *inp2, fpe_internal *result)
{
	int err;
	fpe_internal m;

	err = fpe_fneg(inp2, &m);
	if (err) return err;
	return fpe_fadd(inp1, &m, result);
};



/* FGSLMUL - single-precision multiply inp1 by inp2 */
int fpe_fsglmul (fpe_internal *inp1, fpe_internal *inp2, fpe_internal *result)
{
	int err;
	fpe_internal ir;
	fpe_single sgl;

	err = fpe_fmul (inp1, inp2, &ir);
	if (err) return err;
	err = fpe_internal_to_single (&ir, &sgl);
	if (err) return err;
	return fpe_single_to_internal (&sgl, result);
};



/* FSGLDIV - single-precision divide inp1 by inp2 */
int fpe_fsgldiv (fpe_internal *inp1, fpe_internal *inp2, fpe_internal *result)
{
	int err;
	fpe_internal ir;
	fpe_single sgl;

	err = fpe_fdiv (inp1, inp2, &ir);
	if (err) return err;
	err = fpe_internal_to_single (&ir, &sgl);
	if (err) return err;
	return fpe_single_to_internal (&sgl, result);
};



/* FREM - floating-point remainder */
int fpe_frem (fpe_internal *inp, fpe_internal *mod, fpe_internal *result)
{
	int err;
	fpe_internal ir1, ir2;
	
	err = fpe_fdiv (inp, mod, ir1);
	if (err) return err;
	err = fpe_fintrz (ir1, ir2);
	if (err) return err;
	err = fpe_fmul (ir2, mod, ir1);
	if (err) return err;
	return fsub (inp, ir1, result);
};



/* FMOD - floating point modulo */
/* There are subtle differences between this and FREM, but for now we */
/*     take the easy way out and just define this to be FREM.         */
int fpe_fmod (fpe_internal *inp, fpe_internal *mod, fpe_internal *result)
{
	return fpe_frem (inp, mod, result);
};


/* FCOS - floating-point cosine */
int fpe_fcos (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal ir;

	err = fpe_fadd (inp, FPE_MINUS_PI_OVER_TWO, &ir);
	if (err) return err;
	return fpe_fsin (&ir, result);
};



/* FSINCOS - obtain sine (in result1) and cosine (in result2) of inp */
int fpe_fsincos (fpe_internal *inp, fpe_internal *result1, fpe_internal *result2)
{
	int err;
	fpe_internal ir, res1, res2;

	err = fpe_fsin (inp, &res1);
	if (err) return err;
	err = fpe_fadd (inp, FPE_MINUS_PI_OVER_TWO, &ir);
	if (err) return err;
	err = fpe_fsin (ir, &res2);
	if (err) return err;
	memcpy (result1, &res1, sizeof(fpe_internal));
	memcpy (result2, &res2, sizeof(fpe_internal));
	return 0;
};



/* FTAN - floating-point tangent */
int fpe_ftan (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal ir, ir1, ir2;

	err = fpe_fsin (inp, &ir1);
	if (err) return err;
	err = fpe_fadd (inp, FPE_MINUS_PI_OVER_TWO, &ir);
	if (err) return err;
	err = fpe_fsin (&ir, &ir2);
	if (err) return err;

	/* XXX - what do we do about division by zero error? */
	return fpe_fdiv (&ir1, &ir2, result);
};



/* FASIN - circular arcsine */ 
/* FASIN(x) = FATAN(FDIV(x, FSQRT(FMUL(FADD([1],x), FADD([-1],x))))) */
int fpe_fasin (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal ir1, ir2, ir3;

	err = fpe_fadd (inp, FPE_MINUS_ONE, &ir1);
	if (err) return err;
	err = fpe_fadd (inp, FPE_ONE, &ir2);
	if (err) return err;
	err = fpe_fmul (&ir1, &ir2, &ir3);
	if (err) return err;
	err = fpe_fsqrt (&ir3, &ir1);
	if (err) return err;
	err = fpe_fdiv (inp, &ir1, &ir2);
	if (err) return err;
	return fpe_fatan (&ir2, result);
};



/* FSINH - hyperbolic sine */
/* FSINH(x) = sgn(x)*(1/2)*FADD(z,FDIV(z,FADD([1],z))), where z=FETOXM1(FABS(x)). */
int fpe_fsinh (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal z, ir1, ir2;

	err = fpe_fabs (inp, &ir1);
	if (err) return err;
	err = fpe_fetoxm1 (&ir1, &z);
	if (err) return err;
	err = fpe_fadd (&z, FPE_ONE, &ir1);
	if (err) return err;
	err = fpe_fdiv (&z, &ir1, &ir2);
	if (err) return err;
	err = fpe_fadd (&z, &ir2, &ir1);
	if (err) return err;
	err = fpe_fmul (&ir1, FPE_ONE_HALF, result);
	if (err) return err;
	result->sign *= inp->sign;  /* XXX - the right way? */
	return 0;
};



/* FCOSH - hyperbolic cosine */
/* FCOSH(x) = (1/2)*FADD(FETOX(y),FETOX(-y)), where y=FABS(x). */
int fpe_fcosh (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal y, ir1, ir2, ir3;

	err = fpe_fabs (inp, &y);
	if (err) return err;
	err = fpe_fetox (&y, &ir1);
	if (err) return err;
	err = fpe_fneg (&y, &ir2);
	if (err) return err;
	err = fpe_fetox (&ir2, &ir3);
	if (err) return err;
	err = fpe_fadd (&ir1, &ir3, &ir2);
	if (err) return err;
	return fpe_fmul (&ir2, FPE_ONE_HALF, result);
};



/* FLOGNP1 - logarithm base e of (inp + 1) */
int fpe_flognp1 (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal ir;

	err = fpe_fadd (inp, FPE_ONE, &ir);
	if (err) return err;
	return fpe_flogn (&ir, result);
};



/* FETOXM1 - the exponential function evaluated at x, minus one */
int fpe_fetoxm1 (fpe_internal *inp, fpe_internal *result)
{
	int err;
	fpe_internal ir;

	err = fpe_etox (inp, &ir);
	if (err) return err;
	return fadd (&ir, FPE_MINUS_ONE, result);
};

