#include "internal.h"





/* 
 *  XXX - things I haven't done yet:
 *
 *  - real exception processing
 *  - distinguishing between signaling and non-signaling NAN's
 *
 *  Pipe dreams:
 *  - accounting for bias in single/double/extended exponents
 *    (just subtract off some value)
 *  - different rounding modes
 *    (only round-to-nearest is supported now - also support round-to-zero,
 *    round-to-minus-infinity, round-to-plus-infinity)
 *  - correct "tie-case" (i.e. 0.5) handling in round-to-nearest mode
 *    (we always round upwards, contrary to the IEEE spec)
 */



void fpe_normalize (fpn_internal *f)
{
	if ((f->mantissa) && (f->class == 0))
		while (!(f->mantissa & 0x80000000L) && 
			(f->exponent > -2147483648)) {
			f->mantissa <<=1;
			f->exponent--;
		};
};


void fpe_sint_to_internal (int *i, fpn_internal *f)
{
	unsigned long j;
	int k = 31;

	memset((void *)f, 0, sizeof(fpn_internal));

	if (!(*i & 0x7FFFFFFFL)) {
		if (*i & 0x80000000L) {
			/* largest possible negative value */
			f->sign = -1;
			f->mantissa = 0x80000000000000000L;
			f->exponent = 31;
			return;
		} else
			/* zero */
			return;
	};

	if (*i & 0x80000000L) {
		f->sign = -1;
		j = -*i;
	} else {
		f->sign = 1;
		j = *i;
	};

	while (1) {
		j <<= 1; k--;
		if (j & 0x80000000L) {
			f->mantissa = (j << 32);
			f->exponent = k;
			return;
		};
	};
};



void fpe_internal_to_sint (fpn_internal *f, int *i)
{
	unsigned int r;

	if (f->exponent < 0) {
		/* underflow */
		*i = 0;
		return;
	} else if (f->exponent > 30) {
		/* overflow */
		*i = (f->sign < 0) ? 0x80000000L : 0x7FFFFFFFL;
		return;
	} else if (f->class) {
		/* XXX - NaN, infinity, something else... */
		*i = 0;
		return;
	};
	*i = f->mantissa >> (63 - f->exponent)

	/* round off */
	r = f->mantissa << (33 + f->exponent);
	if ((r & 0x80000000L) && ((*i & 0x7FFFFFFFL) != 0x7FFFFFFFL))
		*i += f->sign;
	return;
};



void fpe_uint_to_internal (unsigned int *i, fpn_internal *f)
{
	unsigned int j;
	int k = 31;

	memset((void *)f, 0, sizeof(fpn_internal));

	if (!(*i)) {
		/* zero */
		return;
	};

	f->sign = 1;
	j = *i;
	while (1) {
		if (j & 0x80000000L) {
			f->mantissa = (j << 32);
			f->exponent = k;
			return;
		};
		j <<= 1; k--;
	};
};



void fpe_internal_to_uint (fpn_internal *f, unsigned int *i)
{
	unsigned int r;

	if (f->sign == -1) {
		/* underflow (?) */
		*i = 0;
		return;
	} else if (f->exponent < 0) {
		/* underflow */
		*i = 0;
		return;
	} else if (f->exponent > 31) {
		/* overflow */
		*i = 0xFFFFFFFFL;
		return;
	} else if (f->class) {
		/* XXX - NaN, infinity, something else... */
		*i = 0;
		return;
	};
	*i = f->mantissa >> (63 - f->exponent);

	/* round off */
	r = f->mantissa >> (31 - f->exponent);
	if ((r & 0x80000000L) && (*i != 0xFFFFFFFFL))
		*i++;
	return;
};



void fpe_single_to_internal (fpn_single *s, fpn_internal *f)
{
	memset((void *)f, 0, sizeof(fpn_internal));

	f->sign = s->sign ? -1 : 1;
	if ((s->exponent == 0) && (s->mantissa == 0L)) {
		/* zero */
		return;
	} else if ((s->exponent == 0xFF) && (s->mantissa)) {
		/* NaN */
		f->class = CLASS_NAN;
		return;
	} else if ((s->exponent == 0xFF) && (s->mantissa == 0L)) {
		/* infinity */
		f->class = CLASS_INF;
		return;
	};

	f->exponent = s->exponent;
	f->mantissa = s->mantissa << 41;
	fpe_normalize(f);
	return;
};



void fpe_internal_to_single (fpn_internal *f, fpn_single *s)
{
	unsigned int r;

	memset((void *)s, 0, sizeof(fpn_single));

	s->sign = (f->sign < 0) ? 1 : 0;
	if ((f->mantissa == 0) && (f->exponent == 0)) {
		/*  zero  */
		return;
	} else if (f->class == CLASS_NAN) {
		/*  NaN  */
		s->exponent = 0xFF;
		s->mantissa = 0x007FFFFFL;
		return;
	} else if (f->class == CLASS_INF) {
		/*  infinity  */
		s->exponent = 0xFF;
		return;
	};

	if (f->exponent < -22) {
		/*  underflow - return 0  */
		return;
	} else if (f->exponent < 0) {
		/*  small number - no problem  */
		r = f->mantissa >> (40 - f->exponent);		
		s->mantissa = r >> 1;
		s->exponent = 0;
		/*  round off  */
		if (r & 1L)
			if (s->mantissa != 0xFFFFFFFFL)
				s->mantissa++;
			else if (s->exponent < 0xFE) {
				s->mantissa = 0x00400000L;
				s->exponent++;
			} else {
				/*  avoid overflow - don't round  */
			};
		return;
	} else if (f->exponent > 0xFE) {
		/*  overflow - return infinity  */
		s->exponent = 0xFF;
		return;
	} else {
		/*  normal number  */
		s->mantissa = f->mantissa >> 41;
		s->exponent = f->exponent;
		/*  round off  */
		if (f->mantissa[0] & 0x0000100L)
			if (s->mantissa != 0x007FFFFFL)
				s->mantissa++;
			else if (s->exponent < 0xFE) {
				s->mantissa = 0x00400000L;
				s->exponent++;
			} else {
				/*  avoid overflow - don't round  */
			};
		return;
	};
};



void fpe_double_to_internal (fpn_double *d, fpn_internal *f)
{
	memset((void *)f, 0, sizeof(fpn_internal));

	f->sign = d->sign ? -1 : 1;
	if ((d->exponent == 0) && (d->mantissa == 0)) {
		/* zero */
		return;
	} else if ((d->exponent == 0x3FF) && (d->mantissa0)) {
		/* NaN */
		f->class = CLASS_NAN;
		return;
	} else if ((d->exponent == 0x3FF) && (d->mantissa0 == 0)) {
		/* infinity */
		f->class = CLASS_INF;
		return;
	};

	f->exponent = d->exponent;
	f->mantissa = d->mantissa << 12;
	fpe_normalize(f);
	return;
};



void fpe_internal_to_double (fpn_internal *f, fpn_double *d)
{
	unsigned long long r;

	memset((void *)d, 0, sizeof(fpn_double));

	d->sign = (f->sign < 0) ? 1 : 0;
	if ((f->mantissa == 0) && (f->exponent == 0)) {
		/*  zero  */
		return;
	} else if (f->class == CLASS_NAN) {
		/*  NaN  */
		d->exponent = 0x3FF;
		d->mantissa = 0x000FFFFFFFFFFFFFLL;
		return;
	} else if (f->class == CLASS_INF) {
		/*  infinity  */
		d->exponent = 0x3FF;
		return;
	};

	if (f->exponent < -51) {
		/*  underflow - return 0  */
		return;
	} else if (f->exponent < 0) {
		/*  small number - no problem  */
		r = (f->mantissa << (21 + f->exponent));
		d->mantissa = r >> 1;
		d->exponent = 0;
		/*  round off  */
		if (r & 1LL)
			if (d->mantissa != 0x000FFFFFFFFFFFFFLL)
				d->mantissa++;
			else if (d->exponent < 0x3FE) {
				d->mantissa = 0x0008000000000000LL;
				d->exponent++;
			} else {
				/*  avoid overflow - don't round  */
			};
		return;
	} else if (f->exponent > 0x3FE) {
		/*  overflow - return infinity  */
		d->exponent = 0x3FF;
		return;
	} else {
		/*  normal number  */
		d->mantissa = f->mantissa >> 12;
		d->exponent = 0;
		/*  round off  */
		if (f->mantissa & 0x0000000000000800LL)
			if (d->mantissa != 0x000FFFFFFFFFFFFFLL)
				d->mantissa++;
			else if (d->exponent < 0x3FE) {
				d->mantissa = 0x0008000000000000LL;
				d->exponent++;
			} else {
				/*  avoid overflow - don't round  */
			};
		return;
	};
};



void fpe_extended_to_internal (fpn_extended *e, fpn_internal *f)
{
	memset((void *)f, 0, sizeof(fpn_internal));

	f->sign = e->sign ? -1 : 1;
	if ((e->exponent == 0) && (e->mantissa == 0LL)) {
		/* zero */
		return;
	} else if ((e->exponent == 0x7FFF) && (e->mantissa)) {
		/* NaN */
		f->class = CLASS_NAN;
		return;
	} else if ((e->exponent == 0x7FFF) && (e->mantissa == 0L)) {
		/* infinity */
		f->class = CLASS_INF;
		return;
	};

	f->exponent = e->exponent;
	f->mantissa = e->mantissa;
	fpe_normalize(f);
	return;
};



void fpe_internal_to_extended (fpn_internal *f, fpn_extended *e)
{
	unsigned long long r;

	memset((void *)d, 0, sizeof(fpn_extended));

	e->sign = (f->sign < 0) ? 1 : 0;
	if ((f->mantissa == 0) && (f->exponent == 0)) {
		/*  zero  */
		return;
	} else if (f->class == CLASS_NAN) {
		/*  NaN  */
		e->exponent = 0x7FFF;
		e->mantissa = 0xFFFFFFFFFFFFFFFFLL;
		return;
	} else if (f->class == CLASS_INF) {
		/*  infinity  */
		e->exponent = 0x7FFF;
		return;
	};

	if (f->exponent < -63) {
		/*  underflow - return 0  */
		return;
	} else if (f->exponent < 0) {
		/*  small number - no problem  */

		r = ((unsigned long long) f->mantissa << 
			(32 + f->exponent));
		e->mantissa = r >> 1;
		e->exponent = 0;
		/*  round off  */
		if (r & 1LL)
			if (e->mantissa != 0xFFFFFFFFFFFFFFFFLL)
				e->mantissa++;
			else if (e->exponent < 0x7FFE) {
				e->mantissa = 0x8000000000000000LL;
				e->exponent++;
			} else {
				/*  avoid overflow - don't round  */
			};
		return;
	} else if (f->exponent > 0x7FFE) {
		/*  overflow - return infinity  */
		e->exponent = 0x7FFF;
		return;
	} else {
		/*  normal number  */
		e->mantissa = f->mantissa;
		e->exponent = f->exponent;
		return;
	};
};









/* 
 *  XXX - packed BCD routines currently lack exception handling. 
 *
 *  (Yes, these functions are fat, slow and ugly.  If it's any consolation, 
 *  very few programs use the packed BCD number format, so we probably won't
 *  notice the speed difference - though the added kernel bloat is still a
 *  problem...)
 */


void fpe_packedbcd_to_internal (fpn_packedbcd *p, fpn_internal *f)
{
	fpn_internal d, t, q;
	int ten;
	unsigned long long digits;

	ten = ((1000*p->exp_3 + 100*p->exp_2 + 10*p->exp_1 + p->exp_0) *
		(p->exponent_sign ? -1 : 1)) - 15;

	/*  Jesus, is this ugly...  */
	digits =  p->digit_0
		+ p->digit_1  * 10LL
		+ p->digit_2  * 100LL
		+ p->digit_3  * 1000LL
		+ p->digit_4  * 10000LL
		+ p->digit_5  * 100000LL
		+ p->digit_6  * 1000000LL
		+ p->digit_7  * 10000000LL
		+ p->digit_8  * 100000000LL
		+ p->digit_9  * 1000000000LL
		+ p->digit_10 * 10000000000LL
		+ p->digit_11 * 100000000000LL
		+ p->digit_12 * 1000000000000LL
		+ p->digit_13 * 10000000000000LL
		+ p->digit_14 * 100000000000000LL
		+ p->digit_15 * 1000000000000000LL
		+ p->digit_16 * 10000000000000000LL;

	memset ((void *)&d, 0, sizeof(fpn_internal));
	d->mantissa[0] = (digits >> 32);
	d->mantissa[1] = digits & 0xFFFFFFFFLL;
	d->exponent = 64;
	d->sign = (p->mantissa_sign ? -1 : 1);

	/* XXX - come back and add error checking */
	fpe_sint_to_internal (&ten, &t);
	fpe_ftentox (&t, &q);
	fpe_fmul (&d, &q, f);
};



void internal_to_packedbcd (fpn_internal *f, fpn_packedbcd *p)
{
/* XXX - pseudocode...
	fpn_internal a, b;
	int power;
	unsigned long long digits, place;

	fpe_flog10 (f, &a);
	fpe_fint (&a, &b);
	fpe_internal_to_sint (&b, &power);
	fpe_fadd (FPE_SIXTEEN, &b, &a);
	fpe_ftentox (&a, &b);
	fpe_fdiv(f, &b, &a);

	/* XXX - 
	... store rounded absolute value of a in the variable digits ...
	... (make sure a->exponent == 63, then:) ...
	*/
	digits = a->mantissa;

	memset ((void *)&p, 0, sizeof(fpn_packedbcd));

	p->digit_0  = digits % 10LL;	digits /= 10LL;
	p->digit_1  = digits % 10LL;	digits /= 10LL;
	p->digit_2  = digits % 10LL;	digits /= 10LL;
	p->digit_3  = digits % 10LL;	digits /= 10LL;
	p->digit_4  = digits % 10LL;	digits /= 10LL;
	p->digit_5  = digits % 10LL;	digits /= 10LL;
	p->digit_6  = digits % 10LL;	digits /= 10LL;
	p->digit_7  = digits % 10LL;	digits /= 10LL;
	p->digit_8  = digits % 10LL;	digits /= 10LL;
	p->digit_9  = digits % 10LL;	digits /= 10LL;
	p->digit_10 = digits % 10LL;	digits /= 10LL;
	p->digit_11 = digits % 10LL;	digits /= 10LL;
	p->digit_12 = digits % 10LL;	digits /= 10LL;
	p->digit_13 = digits % 10LL;	digits /= 10LL;
	p->digit_14 = digits % 10LL;	digits /= 10LL;
	p->digit_15 = digits % 10LL;	digits /= 10LL;
	p->digit_16 = digits % 10LL;
	p->mantissa_sign = (a->sign < 0) ? 1 : 0;

	if (power < 0) {
		p->exponent_sign = 1;
		power = -power;		/* is abs() defined in kernel? */
	};
	p->exp_0 = power % 10;	power /= 10;
	p->exp_1 = power % 10;	power /= 10;
	p->exp_2 = power % 10;	power /= 10;
	p->exp_3 = power % 10;
*/
};
