/* 
 *  arch/m68k/math-emu/fpe_primitives.c
 *  (c) 1997 Rob Pelkey <rpelkey@bates.edu>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *
 *  This file implements the primitive floating-point functions in the
 *  package.  Some could be implemented better, but it's not like we're
 *  relying on this for anything mission-critical - we have -msoft-float
 *  for that. :-)
 *
 *  Change history:
 *  xx Dec 1997 - initial version.
 */

#include "fpe_types.h"


/*
primitives:
* FADD
* FDIV
* FMUL
* FABS	= x->sign=1
* FNEG	= x->sign= -(x->sign)
* FSCALE	= x->exponent += y
* FGETEXP	= x->exponent
* FGETMAN	= x->mantissa
* FINT
* FINTRZ
FSIN	= <<polynomial>>
FETOX
FLOGN
FSQRT
FATAN	= <<polynomial>>
*/



/*  FABS - absolute value  */
int fpe_fabs (fpe_internal *inp, fpe_internal *result)
{
	/* XXX - perform exception processing */

	fpe_copy_normalize (inp, result);
	result->sign = 1;
	return 0;
};



/*  FNEG - negate  */
int fpe_fabs (fpe_internal *inp, fpe_internal *result)
{
	/* XXX - perform exception processing */

	fpe_copy_normalize (inp, result);
	result->sign *= -1;
	return 0;
};



/*  FSCALE - scale base two exponent  */
inf fpe_fscale (fpe_internal *inp, fpe_internal *sc, fpe_internal *result)
{
	fpe_internal res;
	signed int scalar;

	/* XXX - perform exception processing */

	fpe_copy_normalize (inp, &res);
	fpe_internal_to_sint (sc, &scalar);
	res.exponent += scalar;

	/* XXX - perform exception processing */

	memcpy((void *)result, (void *)&res, sizeof(fpe_internal));
	return 0;
};



/*  FGETEXP - get exponent  */
int fpe_fgetexp (fpe_internal *inp, fpe_internal *result)
{
	signed int exp;
	
	/* XXX - perform exception processing */
	
	exp = inp->exponent;
	fpe_sint_to_internal (exp, result);
	return 0;
};



/*  FGETMAN - get mantissa */
int fpe_fgetman (fpe_internal *inp, fpe_internal *result)
{
	/* XXX - perform exception processing */
	
	fpe_copy_normalize (inp, result);
	result->exponent = 0;
	return 0;
};



/*  FINTRZ - integer part, rounded towards zero  */
int fpe_fintrz (fpe_internal *inp, fpe_internal *result)
{
	/* XXX - perform exception processing */
	
	fpe_copy_normalize (inp, result);
	if (result->exponent < 0)
		result->mantissa = 0LL;
	else if (result->exponent < 63)
		result->mantissa &= 0xFFFFFFFFFFFFFFFF << (63 - result->exponent);
	return 0;
};

/*  FINT - integer part  */
/*  The output of this depends on the rounding mode, which we don't  */
/*      implement yet... for now we alias this to FINTRZ.            */
int fpe_fint (fpe_internal *inp, fpe_internal *result)
{
	return fpe_fintrz (inp, result);
};



/*  FADD - floating-point add inp1 and inp2  */
int fpe_fadd (fpe_internal *inp1, fpe_internal *inp2, fpe_internal *result)
{
	fpe_internal li, gi, res, *lx, *gx;
	unsigned long long bitz;
	int bit;

	/* XXX - do exception checking */

	if (inp1->exponent > inp2->exponent) {
		fpe_copy_normalize (inp1, &gi);
		fpe_copy_normalize (inp2, &li);
	} else {
		fpe_copy_normalize (inp1, &li);
		fpe_copy_normalize (inp2, &gi);
	};

	if (gi.exponent - li.exponent < 64)
		li.mantissa >> (gi.exponent - li.exponent);
	else {
		memcpy ((void *)result, (void *)&gi, sizeof(fpe_internal));
		return 0;  /* right? */
	};
	
	if (gi.mantissa > li.mantissa) {
		gx = &li; lx = &gi;
	} else {
		gx = &gi; lx = &li;
	};
	
	if (gx->sign != lx->sign) {
		gx->mantissa -= lx->mantissa;
	} else {
		bitz = (gx->mantissa >> 1) + (lx->mantissa >> 1);
		bit = (gx->mantissa & 1LL) + (lx->mantissa & 1LL);
		gx->exponent++;
		if (bitz & 0x8000000000000000LL)
			gx->mantissa = bitz + (bit >> 1);
		else {
			gx->mantissa = (bitz << 1) + bit;
			gx->exponent--;
		};
	};
	
	/* XXX - do exception checking */

	memcpy ((void *)result, (void *)gx, sizeof(fpe_internal)); 	
	return 0;
};



/*
 *  FMUL - floating-point multiply inp1 by inp2
 *  yes, I'm sure there's a faster way... is bitwise shift/add faster?
 */
int fpe_fmul (fpe_internal *inp1, fpe_internal *inp2, fpe_internal *result)
{
	unsigned long long bitz1, bitz2;
	unsigned short *is1, *is2;
	fpe_internal inpn1, inpn2, res;

	/* XXX - do exception checking */

	fpe_copy_normalize (inp1, &inpn1);
	fpe_copy_normalize (inp2, &inpn2);
	is1 = (unsigned short *)(&(inpn1.mantissa));
	is2 = (unsigned short *)(&(inpn2.mantissa));

	res.sign = inpn1.sign * inpn2.sign;
	res.exponent = inpn1.exponent + inpn2.exponent + 1;

	bitz1 = (inpn1.mantissa & 0x00000000FFFFFFFFLL) *
		(inpn2.mantissa & 0x00000000FFFFFFFFLL);
	bitz1 >>= 32;
	bitz1 += (is1[3] * is2[1] + is1[1] * is2[3]);
	bitz1 >>= 16;
	bitz1 += (is1[3] * is2[0] + is1[2] * is2[1] + is1[1] * is2[2] + 
		is1[0] * is2[3]);
	bitz2 = is1[2] * is2[0] + is1[0] * is2[2];
	bitz2 += (inpn1.mantissa >> 32) * (inpn2.mantissa >> 32);
	if (!(bitz2 & 0x8000000000000000LL)) {
		res.mantissa = (bitz2 << 1) + (bitz1 >> 15);
		res.exponent--
	} else {
		res.mantissa = (bitz2 << 0) + (bitz1 >> 16);	
	};

	/* XXX - do exception checking again */
	
	memcpy((void *)result, (void *)&res, sizeof(fpe_internal));
	return 0;
};



/*  FDIV - floating-point divide inp1 by inp2  */
/*  Thanks to Knuth (vol. 2, 4.2.3) for the algorithm.  */
int fpe_fdiv (fpe_internal *inp, fpe_internal *div, fpe_internal *result)
{
	unsigned long long bitz1, bitz2;
	unsigned long *inl, *dvl, bitl[3];
	fpe_internal ini, dvi, ir, res;

	/* XXX - do exception checking */

	fpe_copy_normalize (inp, &ini);
	fpe_copy_normalize (div, &dvi);
	inl = (unsigned short *)(&(ini.mantissa));
	dvl = (unsigned short *)(&(dvi.mantissa));
	irl = (unsigned short *)(&(ir.mantissa));

	ir.sign = ini.sign * dvi.sign;

	bitz1 = inp.mantissa / dvl[0];
	bitl[0] = bitz1 >> 32;
	bitl[1] = bitz1 & 0x00000000FFFFFFFFLL;
	bitz2 = (inp.mantissa % dvl[0]) << 32;
	bitz1 = bitz2 / dvl[0];
	bitz2 = bitl[1] + (bitz1 >> 32);
	bitl[1] = bitz2 & 0x00000000FFFFFFFFLL;
	if (bitz2 & 0xFFFFFFFF00000000LL) bitl[0] += (bitz2 >> 32);
	bitl[2] = bitz1 & 0x00000000FFFFFFFFLL;
	
	if (bitl[0] & 0x00000001L) {
		ir.exponent = ini.exponent - dvi.exponent;
		ir.mantissa = 0x8000000000000000LL | (bitl[1] << 31) | (bitl[2] >> 1);
	} else {
		ir.exponent = ini.exponent - dvi.exponent - 1;
		ir.mantissa = (bitl[1] << 32) | bitl[2];
	};

	bitz1 = irl[0] * dvl[1] / dvl[0];
	ir.mantissa -= (bitz1 >> 32);
	
	/* XXX - more exception processing */
	
	return 0;
};
